% 
% function run_cppca_pr()
% 
%   This function runs the CPPCA-PR algorithm.
%

%
% CPPCA-PR: CPPCA with Partitioned-Residual Reconstruction
% Copyright (C) 2012  James E. Fowler
% http://www.ece.mstate.edu/~fowler
% 
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

function run_cppca_pr()

addpath(genpath('../CPPCA-3.3-1'));
addpath(genpath('../BCS/MT_CS_demo'));

% num_rows = 64;
% num_cols = 64;
% N = 169;
% filename = 'hydice';
% Ls_normal = [10 10 10 10 10 10];
% Ls_anomaly = [3 3 3 3 3 3];
% anomaly_percentage = 0.00463;
% gamma_factor = 100;

num_rows = 200;
num_cols = 200;
N = 158;
filename = 'lcvf';
Ls_normal = [3 5 6 9 12 16];
Ls_anomaly = [1 3 3 5 5 5];
anomaly_percentage = 2/num_rows/num_cols;
gamma_factor = 1;

M = num_rows * num_cols;

fid = fopen([filename '.dat'], 'r', 'ieee-be');
X = fread(fid, [N M], 'float32');
fclose(fid);

fid = fopen([filename '.ground_truth.dat'], 'r', 'ieee-be');
ground_truth_map2 = fread(fid, [1 M], 'uint8');
fclose(fid);

[N M] = size(X);
J = 20;
projection_matrix_file = ['projections.' num2str(N) '.' num2str(J) '.mat'];
lifting_scheme = liftwave('db4');
num_levels = 4;
Psi_D4 = DWT_Matrix(N, lifting_scheme, num_levels);

relative_dimensions = [0.1 0.15 0.2 0.3 0.4 0.5];

Ks = round(relative_dimensions * N);

X_original = X;

for j = 1:J
  anomaly_map{j} = logical(ground_truth_map2(j:J:M));
  normal_map{j} = ~anomaly_map{j};
end

clear X
disp('Running Projections...');
for index1 = 1:length(Ks)
  
  K = Ks(index1);
  
  P{index1} = CPPCA_GenerateProjections(N, K, J, projection_matrix_file);

  [Y_tilde{index1}, X{index1}] = CPPCA_Encoder(X_original, P{index1});
end  

disp('Running Projected RX...');
for index1 = 1:length(Ks)
  for j = 1:J
    D{index1}{j} = RX(X{index1}{j});
  end

  tau{index1} = RX_threshold(cell2mat(D{index1}), 1 - ...
      anomaly_percentage);
  for j = 1:J
    anomaly_map_rx{index1}{j} = (D{index1}{j} > tau{index1});
  end
  
  for j = 1:J
    D_projected{index1}{j} = RX(Y_tilde{index1}{j});
  end
  tau_projected{index1} = RX_threshold(cell2mat(D_projected{index1}), ...
      1 - anomaly_percentage);
  for j = 1:J
    anomaly_map_projected{index1}{j} = ...
	(D_projected{index1}{j} > tau_projected{index1});
    normal_map_projected{index1}{j} = ...
	(~anomaly_map_projected{index1}{j});
  end
  write_projected_map(anomaly_map_projected{index1}, num_rows, ...
      num_cols, relative_dimensions(index1), filename);
  disp(['Percent correct (K/N = ' num2str(relative_dimensions(index1)) ...
	'): ' num2str((M - ...
	sum(xor(cell2mat(anomaly_map_projected{index1}), ...
	cell2mat(anomaly_map)))) / M * 100)]);
end

disp('Partitioning dataset...');
for index1 = 1:length(Ks)
  for j = 1:J
    Y_tilde_normal{index1}{j} = ...
	Y_tilde{index1}{j}(:, normal_map_projected{index1}{j});
    Y_tilde_anomaly{index1}{j} = ...
	Y_tilde{index1}{j}(:, anomaly_map_projected{index1}{j});
    X_normal{index1}{j} = X{index1}{j}(:, normal_map_projected{index1}{j});
    X_anomaly{index1}{j} = X{index1}{j}(:, anomaly_map_projected{index1}{j});
  end
end

disp('Running CPPCA on normal pixels...');
for index1 = 1:length(Ks)
  K = Ks(index1);
  L = Ls_normal(index1);

  tic
  [X_check_normal{index1}, Psi_CPPCA{index1}] = ...
      CPPCA_Decoder(Y_tilde_normal{index1}, P{index1}, L);
  for j = 1:J
    X_hat_normal{index1}{j} = ...
	[Psi_CPPCA{index1} zeros(N, N - L)] * ...
	X_check_normal{index1}{j};
  end
  toc
end

disp('Running MT-BCS-D4 on anomaly-pixel residuals...');
for index1 = 1:length(Ks)
  K = Ks(index1);
  L = Ls_anomaly(index1);

  tic
  if (L > 0)
    Psi_CPPCA{index1} = Psi_CPPCA{index1}(:, 1:L);
    X_check_CPPCA = GetCoefficients(Y_tilde_anomaly{index1}, ...
	P{index1}, Psi_CPPCA{index1});
    for j = 1:J
      X_CPPCA_anomaly{j} = [Psi_CPPCA{index1} zeros(N, N - L)] * ...
	  X_check_CPPCA{j};
      Y_p = P{index1}{j}' * X_CPPCA_anomaly{j};
      R{j} = Y_tilde_anomaly{index1}{j} - Y_p;
    end
    R_check = MTBCS_Decoder(R, P{index1}, Psi_D4, gamma_factor);
    for j = 1:J
      X_hat_anomaly{index1}{j} = X_CPPCA_anomaly{j} + Psi_D4 * R_check{j};
    end
  else
    X_check = MTBCS_Decoder(Y_tilde_anomaly{index1}, P{index1}, Psi_D4);
    for j = 1:J
      X_hat_anomaly{index1}{j} = Psi_D4 * X_check{j};
    end
  end
  toc

end

for index1 = 1:length(Ks)
  K = Ks(index1);
  
  for j = 1:J
    X_hat{j} = zeros(N, length(anomaly_map_projected{index1}{j}));
    X_hat{j}(:, anomaly_map_projected{index1}{j}) = ...
	X_hat_anomaly{index1}{j};
    X_hat{j}(:, normal_map_projected{index1}{j}) = ...
	X_hat_normal{index1}{j};
  end

  D_PR_normal(index1) = SNR_PartitionedDataset(X{index1}, X_hat, ...
      normal_map, eye(N));
  S_PR_normal(index1) = SAM_PartitionedDataset(X{index1}, X_hat, ...
      normal_map, eye(N));
  D_PR_anomaly(index1) = SNR_PartitionedDataset(X{index1}, X_hat, ...
      anomaly_map, eye(N));
  S_PR_anomaly(index1) = SAM_PartitionedDataset(X{index1}, X_hat, ...
      anomaly_map, eye(N));
end

save([filename '.partitioned_residual.results.mat'], ...
      'relative_dimensions', ...
      'D_PR_normal', 'S_PR_normal', ...
      'D_PR_anomaly', 'S_PR_anomaly');

disp(['PR normal: ' num2str(D_PR_normal) ' dB']);
disp(['PR anomaly: ' num2str(D_PR_anomaly) ' dB']);
disp(['PR normal: ' num2str(S_PR_normal) ' degrees']);
disp(['PR anomaly: ' num2str(S_PR_anomaly) ' degrees']);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function X_check = GetCoefficients(Y_tilde, P, Psi)

J = length(Y_tilde);

L = size(Psi, 2);

for j = 1:J
  [K Mj] = size(Y_tilde{j});
  [N K] = size(P{j});
  X_check{j} = zeros(N, Mj);

  X_check{j}(1:L,:) = pinv(P{j}' * Psi) * Y_tilde{j};
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function write_projected_map(anomaly_map, num_rows, num_cols, ...
    relative_dimension, filename)

J = length(anomaly_map);
M = num_rows * num_cols;
anomaly_map2 = zeros(1, M);
for j = 1:J
  anomaly_map2(j:J:M) = anomaly_map{j};
end
anomaly_map2 = reshape(anomaly_map2, [num_cols num_rows])';
imwrite(uint8(255*anomaly_map2), ...
    [filename '.anomalymap.' num2str(relative_dimension * 100) '.pgm']);

